#include "Event.h"

SLIST EVENT::event_fds;
EVENT_CALLBACK EVENT::rmp_callback = (EVENT_CALLBACK)0;
SLIST EVENT::alarms;

void
EVENT::LOOP() {
  fd_set file_descriptors;
  EVENT_FD* event_fd;
  mRMPEvent* event;

  FOREVER {
    FD_ZERO(&file_descriptors);
    
    SLIST_ITERATOR iter(event_fds);
    if (MTCP_SERVER::TerminateQ() == 0) {
      while ( event_fd = (EVENT_FD*)++iter ) {
	FD_SET(event_fd->FD(), &file_descriptors);
      }
    }
    
    event = mRMP::mRMPLoop(FD_SETSIZE,&file_descriptors,NULL,NULL,NULL);

    // Handle RMP Events.
    if (event != NULL) rmp_callback((CLIENTDATA)event);

    // Handle file descriptor events.
    if (MTCP_SERVER::TerminateQ() == 0) {
      iter.reset();
      while ( event_fd = (EVENT_FD*)++iter ) {
	if (FD_ISSET(event_fd->FD(), &file_descriptors)) {
	  event_fd->CALL();
	  break;
	}
      }
    }
  }
}

void
EVENT::Add_FD_CB(int _FD, EVENT_CALLBACK _CB, CLIENTDATA _D) {
  event_fds.append(new EVENT_FD(_FD, _CB, _D));
}

void
EVENT::Remove_FD_CB(int _FD) {
  EVENT_FD* event_fd;
  SLIST_ITERATOR iter(event_fds);
  while ( event_fd = (EVENT_FD*)++iter ) {
    if (_FD == event_fd->FD()) {
      iter.destroy();
      return;
    }
  }
}

ALARM*
EVENT::SetAlarmOnce(unsigned long _timer_duration, EVENT_CALLBACK _callback, CLIENTDATA _data) {
  ALARM* alarm = new ALARM(ALARM_ONCE, _timer_duration, _callback, _data);
  alarms.append(alarm);
  return alarm;
}

ALARM*
EVENT::SetAlarmMany(unsigned long _timer_duration, EVENT_CALLBACK _callback, CLIENTDATA _data) {
  ALARM* alarm = new ALARM(ALARM_MANY, _timer_duration, _callback, _data);
  alarms.append(alarm);
  return alarm;
}

ALARM*
EVENT::SetAlarmCheck(unsigned long _timer_duration, EVENT_CALLBACK _callback, CLIENTDATA _data) {
  ALARM* alarm = new ALARM(ALARM_CHECK, _timer_duration, _callback, _data);
  alarms.append(alarm);
  return alarm;
}

void
EVENT::RemoveAlarm(ALARM*& _alarm_to_remove) {
  if (_alarm_to_remove != 0) {
    ALARM* alarm;
    SLIST_ITERATOR iter(alarms);
    while (alarm = (ALARM*)++iter) {
      if (_alarm_to_remove == alarm) {
	alarm->Off();
	iter.destroy();
	break;
      }
    }
  }
  _alarm_to_remove = (ALARM*)0;
}

ALARM::ALARM(unsigned long _alarm_type,
	     unsigned long _timer_duration,
	     EVENT_CALLBACK _callback,
	     CLIENTDATA _data)
: alarm_type(_alarm_type),
  timer_duration(_timer_duration),
  callback(_callback),
  data(_data)
{
  switch (_alarm_type) {
  case ALARM_ONCE:
    RMP::setTimer(MAX2(_timer_duration,1),2,(CLIENTDATA)this, ALARM::Once);
    break;
  case ALARM_MANY:
    RMP::setPeriodicAlarm(MAX2(_timer_duration,20),2,(CLIENTDATA)this, ALARM::Many);
    break;
  case ALARM_CHECK:
    TIME_GET(&start);
    RMP::setTimer(MAX2(_timer_duration/1000,1),2,(CLIENTDATA)this, ALARM::Check);
  }
}

void
ALARM::Off() {
  switch(alarm_type) {
  case ALARM_ONCE:
    RMP::cancelAlarm(2,(CLIENTDATA)this, ALARM::Once);
    break;
  case ALARM_MANY:
    RMP::cancelAlarm(2,(CLIENTDATA)this, ALARM::Many);
    break;
  case ALARM_CHECK:
    RMP::cancelAlarm(2,(CLIENTDATA)this, ALARM::Check);
    break;
  }
}

void
ALARM::Once(CLIENTDATA _alarm) {
  ((ALARM*)_alarm)->Ring();
  EVENT::RemoveAlarm((ALARM*&)_alarm);
}

void
ALARM::Many(CLIENTDATA _alarm) {
  ((ALARM*)_alarm)->Ring();
}

void
ALARM::Check(CLIENTDATA _alarm) {
  ALARM* alarm = (ALARM*)_alarm;
  alarm->TimeMarkEnd();
  unsigned long time_expired = alarm->TimeDiff();
  if (time_expired >= alarm->TimerDuration()) {
    alarm->Ring();
    EVENT::RemoveAlarm(alarm);
  } else {
    RMP::setTimer(MAX2((alarm->TimerDuration()-time_expired)/1000,1),2,
		  (CLIENTDATA)alarm, ALARM::Check);
  }
}

void
ALARM::Ring() {
  callback(data);
}
