#include "List.h"

SLIST_NODE::SLIST_NODE() : c(0), n(0) {
}

SLIST_NODE::SLIST_NODE(COLLECTABLE* _c, SLIST_NODE* _n) : c(_c), n(_n) {
}

SLIST::SLIST() : h(0), t(0), e(0) {
}

SLIST::~SLIST() {
  destroyAll();
}

COLLECTABLE*
SLIST::append(COLLECTABLE* i) {
  if (e++) {
    t->n = new SLIST_NODE(i);
    t = t->n;
  } else {
    t = h = new SLIST_NODE(i);
  }
  return i;
}

void
SLIST::removeAll() {
  while ( removeFirst() );
}

void
SLIST::destroyAll() {
  COLLECTABLE* tmp;
  while ( tmp = removeFirst() ) delete tmp;
}

COLLECTABLE*
SLIST::insert(COLLECTABLE* i) {
  if (e) {
    SLIST_ITERATOR iter(*this);
    while ( COLLECTABLE* c = ++iter ) {
      if (i->CMP(c) >= 0) { // if (c >= i)
	return iter.insertBefore(i);
      }
    }
  }
  return append(i);
}

COLLECTABLE*
SLIST::insertAfter(SLIST_NODE* node, COLLECTABLE* i) {
  if (node) {
    SLIST_NODE* handle = new SLIST_NODE(i, node->n);
    if (t == node) t = handle;
    node->n = handle;
    e++;
  } else {
    return prepend(i);
  }
  return i;
}

COLLECTABLE*
SLIST::prepend(COLLECTABLE* i) {
  if (e++) {
    h = new SLIST_NODE(i,h);
  } else {
    t = h = new SLIST_NODE(i);
  }
  return i;
}

COLLECTABLE*
SLIST::removeFirst() {
  if (h) {
    e--;
    SLIST_NODE* tmp = h;
    h = h->n;
    if (!h) t = (SLIST_NODE*)0;
    COLLECTABLE* return_val = tmp->c;
    delete tmp;
    return return_val;
  }
  return (COLLECTABLE*)0;
}

void
SLIST::destroyFirst() {
  delete removeFirst();
}

COLLECTABLE*
SLIST::removeLast() {
  if (t) {
    SLIST_ITERATOR iter(*this);
    while ((++iter) != t->c) ;
    return iter.remove();
  }
  return (COLLECTABLE*)0;
}

void
SLIST::destroyLast() {
  delete removeLast();
}

COLLECTABLE*
SLIST::operator[](int i) {
  if ((i >= e) || (i < 0)) return 0;
  SLIST_NODE* handle = h;
  while(i--) { handle = handle->n; }
  return (handle->c);
}

SLIST_ITERATOR::SLIST_ITERATOR(SLIST& _l) : c(0) {
  l = &_l;
}

SLIST_NODE*
SLIST_ITERATOR::advance_current() {
  clast = c;
  if (c) {
    if (c->n) {
      c = c->n;
    } else {
      c = (SLIST_NODE*)0;
    }
  } else {
    c = l->h;
  }
  if (!c) clast = 0;
  return c;
}

COLLECTABLE* 
SLIST_ITERATOR::operator++() {
  advance_current();
  if (c) {
    return c->c;
  } else {
    return (COLLECTABLE*)0;
  }
}

COLLECTABLE*
SLIST_ITERATOR::operator++(int) {
  COLLECTABLE* return_val;
  if (c) {
    return_val =  c->c;
  } else {
    return_val = (COLLECTABLE*)0;
  }
  advance_current();
  return return_val;
}

COLLECTABLE*
SLIST_ITERATOR::operator()() {
  if (c) {
    return c->c;
  } else {
    return (COLLECTABLE*)0;
  }
}

COLLECTABLE*
SLIST_ITERATOR::findNext(const COLLECTABLE* k) {
  while ( advance_current() ) if ( (c->c)->CMP(k) == 0 ) return c->c;
  return (COLLECTABLE*)0;
}

void
SLIST_ITERATOR::reset() {
  c = (SLIST_NODE*)0;
}

COLLECTABLE*
SLIST_ITERATOR::insertAfter(COLLECTABLE*i) {
  return l->insertAfter(c,i);
}

COLLECTABLE*
SLIST_ITERATOR::insertBefore(COLLECTABLE*i) {
  return l->insertAfter(clast,i);
}

COLLECTABLE*
SLIST_ITERATOR::remove() {
  if (c) {
    SLIST_NODE* next;
    l->e--;
    if (l->t == c) l->t = clast;
    if (l->h == c) {
      next = l->h = l->h->n;
    } else {
      next = clast->n = c->n;
    }
    COLLECTABLE* return_val = c->c;
    delete c;
    c = next;
    return return_val;
  }
  return (COLLECTABLE*)0;
}

void
SLIST_ITERATOR::destroy() {
  delete remove();
}
