#ifndef LIST_H
#define LIST_H

#include "General.h"
#include "Collectable.h"
#include "Collection.h"
#include "Iterator.h"

class SLIST;
class SLIST_ITERATOR;

class SLIST_NODE {
  friend SLIST;
  friend SLIST_ITERATOR;
private:
  COLLECTABLE* c; // Contents.
  SLIST_NODE* n; // Next.
  SLIST_NODE();
  SLIST_NODE(COLLECTABLE*, SLIST_NODE* = 0);
public:
};

class SLIST : public COLLECTION {
  friend SLIST_ITERATOR;
protected:
  SLIST_NODE* h; // Head.
  SLIST_NODE* t; // Tail.
  int e; // Number of Entries.
public:
  SLIST();
  virtual ~SLIST();

  virtual COLLECTABLE* append(COLLECTABLE*);
  virtual int          entries() const { return e; }
  virtual COLLECTABLE* first() const { if (h) return h->c; else return (COLLECTABLE*)0; }
  virtual COLLECTABLE* insert(COLLECTABLE*);
  virtual COLLECTABLE* insertAfter(SLIST_NODE*, COLLECTABLE*);
  virtual COLLECTABLE* last() const { if (t) return t->c; else return (COLLECTABLE*)0; }
  virtual COLLECTABLE* prepend(COLLECTABLE* a);
  virtual COLLECTABLE* removeFirst();
  virtual void         destroyFirst();
  virtual COLLECTABLE* removeLast();
  virtual void         destroyLast();
  virtual void         removeAll();
  virtual void         destroyAll();
  virtual COLLECTABLE* operator[](int);
};

class SLIST_ITERATOR : public ITERATOR {
  friend SLIST;
private:
protected:
  SLIST* l; // List.
  SLIST_NODE* c; // Current node.
  SLIST_NODE* clast;
  virtual SLIST_NODE* advance_current();
public:
  SLIST_ITERATOR(SLIST&);
  virtual COLLECTABLE* operator++(); // Prefix.
  virtual COLLECTABLE* operator++(int); // Postfix.
  virtual COLLECTABLE* operator()();
  virtual COLLECTABLE* findNext(const COLLECTABLE*);
  virtual void reset();
  virtual COLLECTABLE* insertAfter(COLLECTABLE*);
  virtual COLLECTABLE* insertBefore(COLLECTABLE*);
  virtual COLLECTABLE* remove();
  virtual void destroy();
};

#endif
