#include "MTCP_SERVER.h"

/* #undef ENABLE_FORWARD_SIGTSTP */

/* MTCP_MESSAGE */

MTCP_MESSAGE::MTCP_MESSAGE() {
  s = new ostrstream;
  msg_data = (char*)0;
}

MTCP_MESSAGE::~MTCP_MESSAGE() {
  delete msg_data;
  delete s;
}

SendStatus
MTCP_MESSAGE::Issue(MTCP_SOCKET* mtcp_socket) {
  if (msg_data == (char*)0) {
    msg_data = S().str();
    msg_length = (unsigned long int)S().pcount();
  }
  return( mtcp_socket->Send(msg_data, msg_length) );
}

/* MTCP_CONNECTION */

MTCP_CONNECTION::MTCP_CONNECTION(int _fd) {
  s = (ostrstream*)0;
  fd = _fd;
  istrstream* msg = Accept();
  (*msg) >> decode(owner_fd);
  int owner_pid;
  while( (*msg) >> decode(owner_pid) ) {
    owner_pids.append(new INTEGER(owner_pid));
  }
  delete msg;

  EVENT::Add_FD_CB(fd, MTCP_SERVER::Handle_IPCMessage, (CLIENTDATA)this);
}

MTCP_CONNECTION::~MTCP_CONNECTION() {
  delete s;
  owner_pids.destroyAll();
}

int
MTCP_CONNECTION::IDEQ(int sockid, int ownerpid) {
  if (sockid == owner_fd) {
    return PIDEQ(ownerpid);
  }
  return FALSE;
}

int
MTCP_CONNECTION::PIDEQ(int ownerpid) {
  COLLECTABLE* foundpid = 0;
  INTEGER searchpid(ownerpid);
  SLIST_ITERATOR iter(owner_pids);
  foundpid = iter.findNext( &searchpid );
  if (foundpid != 0) return TRUE;
  return FALSE;
}

void
MTCP_CONNECTION::AppendPID(int ownerpid) {
  if (PIDEQ(ownerpid) == FALSE) {
    owner_pids.append(new INTEGER(ownerpid));
  }
}

void
MTCP_CONNECTION::Reset() {
  delete s;
  s = new ostrstream;
}

void
MTCP_CONNECTION::Issue() {
  char* str = S().str();
  unsigned long msg_length = htonl(S().pcount());
  writen(fd, (char*)(&msg_length), 4);
  writen(fd, str, S().pcount());
  delete str;
}

istrstream*
MTCP_CONNECTION::Accept() {
  unsigned long msg_length;
  if (readn(fd, (char*)(&msg_length), sizeof(unsigned long)) == 0) exit(0);
  msg_length = ntohl(msg_length);
  char* msg = new char[msg_length];
  if (readn(fd, msg, msg_length) == 0) exit(0);
  return new istrstream(msg,(int)msg_length);
}

/* MTCP_SOCKET */

MTCP_SOCKET::MTCP_SOCKET(int _fd) {
  s = (ostrstream*)0;
  fd = _fd;
  istrstream* msg = MTCP_CONNECTION::Accept();
  (*msg) >> decode(owner_fd);
  int owner_pid;
  while( (*msg) >> decode(owner_pid) ) {
    owner_pids.append(new INTEGER(owner_pid));
  }
  delete msg;
  
  EVENT::Add_FD_CB(fd, MTCP_SERVER::Handle_SocketData, (CLIENTDATA)this);
  
  client.setfd((int)this);
  group = (mRMPGroup*)0;
  message = (MTCP_MESSAGE*)0;
  message_level = 0;
  member = FALSE;
  memset((char*)&rmpaddr, (char)0, sizeof(rmpaddr));
  number_accepts = 0;
}

MTCP_SOCKET::~MTCP_SOCKET() {
  group->leave(); /* Memory freed by release in Handle_RMP_Event. */
}

int
MTCP_SOCKET::Connect(istrstream& msg) {
  msg.read((char*)&rmpaddr, sizeof(rmpaddr));
  if (ntohl(rmpaddr.srmp_unicast) == TRUE) {
    if (ntohl(rmpaddr.srmp_ipaddr) > 0) {
      group = mRMP::join(rmpaddr.srmp_groupname,
			 (unsigned short int)ntohl(rmpaddr.srmp_min_size),
			 (Source)ntohl(rmpaddr.srmp_ipaddr),
			 (unsigned short int)ntohl(rmpaddr.srmp_ipport),
			 &client,
			 1 );
      cerr << "Joining group via non-multicast." << endl;
    } else {
      cerr << "HOST: Non-multicast capable, must specify" << endl;
      cerr << "      IP Address/Port of known member" << endl;
      cerr.flush();
      exit(1);
    }
  } else {
    if (RMP::mcastCapable()) {
      cerr
      << "Joining group: >" << rmpaddr.srmp_groupname
      << "< min_size: " << ntohl(rmpaddr.srmp_min_size)
      << " ttl: " << ntohl(rmpaddr.srmp_ttl)
      << " join_flags: " << ntohl(rmpaddr.srmp_join_flags)
      << endl; cerr.flush();
      group = mRMP::join(rmpaddr.srmp_groupname,
			 (unsigned short int)ntohl(rmpaddr.srmp_min_size),
			 0,
			 0,
			 (unsigned char)ntohl(rmpaddr.srmp_ttl),
			 (int)ntohl(rmpaddr.srmp_join_flags),
			 &client,
			 1 );
    } else {
      cerr << "HOST: Non-multicast capable, must specify" << endl;
      cerr << "      IP Address/Port of known member" << endl;
      cerr.flush();
      exit(1);
    }
  }
  group->flags(PERFORM_REFORMATION);
  return 0;
}

int
MTCP_SOCKET::Close(int pid) {
  int rval = -1;
  INTEGER searchpid(pid);
  SLIST_ITERATOR iter(owner_pids);
  if (iter.findNext(&searchpid) != 0) {
    iter.destroy();
    rval = 0;
  }
  if (owner_pids.entries() == 0) {
    close(fd);
    rval = 1;
  }
  return rval;
}

#define SCRATCH_SIZE 8192

void
MTCP_SOCKET::HandleData() {
  static char scratch[SCRATCH_SIZE];
  int length;
  int message_bounded = TRUE;

  if (message == (MTCP_MESSAGE*)0) {
    message_bounded = FALSE;
    cerr << "WARNING: Message not properly delimited, place beginmessage(), endmessage() around write/send calls." << endl;
    cerr.flush();
    BeginMessage();
  }
  
  set_fl(fd, O_NONBLOCK);

  while ( (length = read(fd, scratch, SCRATCH_SIZE)) > 0) {
    message->S().write(scratch, length);
  }
  
  clr_fl(fd, O_NONBLOCK);
  
  if (message == FALSE) {
    EndMessage();
  }
}

void
MTCP_SOCKET::HandleMessage(char* msg_data, int msg_length) {
  write(fd, msg_data, msg_length);
}

void
MTCP_SOCKET::Member(mRMPGroup* G) {
  if (group == G) {
    member = TRUE;
    SendQ();
  }
}

void
MTCP_SOCKET::SendQ() {
  MTCP_MESSAGE* send_message;
  SLIST_ITERATOR iter(message_queue);
  while( send_message = (MTCP_MESSAGE*)++iter ) {
    send_message->Issue(this);
  }
  message_queue.destroyAll();
}

void
MTCP_SOCKET::BeginMessage() {
  if (message_level == 0) {
    delete message;
    message = new MTCP_MESSAGE;
  }
  message_level++;
}

void
MTCP_SOCKET::EndMessage() {
  message_level--;
  if (message_level < 0) return;
  if (message_level == 0) {
    if (member == TRUE) {
      message->Issue(this);
      delete message;
    } else {
      message_queue.append(message);
    }
    message = (MTCP_MESSAGE*)0;
  }
}

SendStatus
MTCP_SOCKET::Send(char* msg_data, unsigned long int msg_length) {
  return( group->send(msg_data, msg_length, (QOS)ntohl(rmpaddr.srmp_qos), 0, FALSE, 1, FALSE) );
}

void
MTCP_SOCKET::GetSockName(ostream& s) {
  if (group == (mRMPGroup*)0) {
    s << encode((int)-1); // Not connected - Error.
  } else {
    s << encode((int)0); // Success.
    s << encode((int)sizeof(rmpaddr));
    s.write((char*)&rmpaddr, sizeof(rmpaddr));
  }
}

int
MTCP_SOCKET::SetSockOpt(istream& msg) {
  int optname;
  int optlen;
  msg >> decode(optname);
  msg >> decode(optlen);
  switch (optname) {
  case SO_QOS:
    {
      if (optlen != sizeof(unsigned long)) break;
      msg.read((char*)&rmpaddr.srmp_qos, optlen);
      return 0;
    }
  default:
    break;
  }
  return -1;
}

void
MTCP_SOCKET::GetSockOpt(istream& msg, ostream& rval) {
  int optname;
  msg >> decode(optname);
  switch (optname) {
  case SO_MINSIZE:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_min_size, sizeof(unsigned long));
    break;
  case SO_TTL:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_ttl, sizeof(unsigned long));
    break;
  case SO_PORT:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_port, sizeof(unsigned long));
    break;
  case SO_IPPORT:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_ipport, sizeof(unsigned long));
    break;
  case SO_IPADDR:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_ipaddr, sizeof(unsigned long));
    break;
  case SO_UNICAST:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_unicast, sizeof(unsigned long));
    break;
  case SO_JOINFLAGS:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_join_flags, sizeof(unsigned long));
    break;
  case SO_QOS:
    rval << encode((int)0);
    rval << encode((int)sizeof(unsigned long));
    rval.write((char*)&rmpaddr.srmp_qos, sizeof(unsigned long));
    break;
  case SO_GROUPNAME:
    rval << encode((int)0);
    rval << encode((int)strlen(rmpaddr.srmp_groupname));
    rval.write(rmpaddr.srmp_groupname, strlen(rmpaddr.srmp_groupname));
    break;
  default:
    break;
  }
  rval << encode((int)-1);
}

void
MTCP_SOCKET::Accept(MTCP_CONNECTION* ipc_connection) {
  if (number_accepts > 0) {
    ipc_connection->Reset();
    ipc_connection->S() << encode(0);
    ipc_connection->Issue();
    number_accepts--;
  }
}

void
MTCP_SOCKET::Bind() {
  number_accepts++;
  Reset();
  S().put(CMD_ACCEPT_REQ);
  Issue();
}

/* MTCP_SERVER */

int MTCP_SERVER::fd;
SLIST MTCP_SERVER::ipc_connections;
SLIST MTCP_SERVER::socket_connections;
CALLBACK_INT MTCP_SERVER::signal_TSTP = (CALLBACK_INT)0;
CALLBACK_INT MTCP_SERVER::signal_CONT = (CALLBACK_INT)0;
int MTCP_SERVER::terminate = 0;
int MTCP_SERVER::exit_status = 0;

void
MTCP_SERVER::Init(int ipcfd) {
  mRMP::init(EVENT_DRIVEN);
  mRMP::setBufferSizeLimits(mRMP_DATA_SIZE,0);
  mRMP::setIncDelivery(FALSE);
  
  InitSignalHandlers();

  MTCP_CONNECTION* ipc_connection = new MTCP_CONNECTION(ipcfd);
  ipc_connections.append(ipc_connection);
  
  /* Create Socket. */
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    cerr << "Error opening MTCP Server socket." << endl; cerr.flush();
    exit(1);
  }
  
  /* Name Socket. */
  struct sockaddr_in server;
  memset((char*)&server, (char)0, sizeof(server));
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  server.sin_port = 0;
  if (bind(fd, (struct sockaddr*)&server, sizeof(server))) {
    cerr << "Error binding MTCP Server socket." << endl; cerr.flush();
    exit(1);
  }
  
  /* Find assigned port number. */
  int length = sizeof(server);
  if (getsockname(fd, (struct sockaddr*)&server, &length)) {
    cerr << "Error getting MTCP Server port." << endl; cerr.flush();
    exit(1);
  }
  unsigned long server_port = (unsigned long)ntohs(server.sin_port);

  /* Start accepting connections. */
  listen(fd, 5);

  /* Tell client our port number. */
  ipc_connection->Reset();
  ipc_connection->S() << encode(server_port);
  ipc_connection->Issue();

  EVENT::RMPEvent_CB(Handle_RMP_Event);
  EVENT::Add_FD_CB(fd, Handle_SocketConnection, (CLIENTDATA)0);
  EVENT::LOOP();
  exit(0);
}

void
MTCP_SERVER::Handle_SocketConnection(CLIENTDATA) {
  int type;
  int FD = accept(fd, 0, 0);
  if (FD == -1) {
    cerr << "Error accepting MTCP IPC connection." << endl; cerr.flush();
    return;
  }
  istrstream* msg = IPCM::Accept(FD);
  (*msg) >> decode(type);
  delete msg;
  switch(type) {
  case CONNECTION_IPC:
    {
      MTCP_CONNECTION* ipc_connection = new MTCP_CONNECTION(FD);
      ipc_connections.append(ipc_connection);
      ipc_connection->Reset();
      ipc_connection->S().put(CMD_IPC_ACK);
      ipc_connection->Issue();
    } break;
  case CONNECTION_SOCKET:
    {
      MTCP_SOCKET* mtcp_socket = new MTCP_SOCKET(FD);
      socket_connections.append(mtcp_socket);
      mtcp_socket->Reset();
      mtcp_socket->S().put(CMD_SOCKET_ACK);
      mtcp_socket->Issue();
    } break;
  }
}

#define FIND_SOCKET 					\
/* Get Args */						\
int sockid;						\
int pid;						\
(*msg) >> decode(sockid);				\
(*msg) >> decode(pid);					\
							\
/* Find Socket */					\
MTCP_SOCKET* socket_connection;				\
SLIST_ITERATOR iter(socket_connections);		\
while ( socket_connection = (MTCP_SOCKET*)++iter ) {	\
  if (socket_connection->IDEQ(sockid, pid)) {		\
    break;						\
  }							\
  socket_connection = (MTCP_SOCKET*)0;			\
}

void
MTCP_SERVER::Handle_IPCMessage(CLIENTDATA _ipc_connection) {
  int cmd_bind = 0;
  MTCP_CONNECTION* ipc_connection = (MTCP_CONNECTION*)_ipc_connection;

  istrstream* msg = ipc_connection->Accept();
  COMMAND cmd;
  msg->get(cmd);
  switch(cmd) {
  case CMD_BIND:
    cmd_bind = 1;
  case CMD_CONNECT:
    {
      /* Return Value */
      int rval = 0;

      FIND_SOCKET
      
      /* Process Connect */
      if (socket_connection != 0) {
	rval = socket_connection->Connect(*msg);
	if (cmd_bind == 1) {
	  socket_connection->Bind();
	}
      } else {
	rval = -1;
      }
      
      /* Return Result */
      ipc_connection->Reset();
      ipc_connection->S() << encode(rval);
      ipc_connection->Issue();
    } break;
  case CMD_CLOSE:
    {
      /* Return Value */
      int rval = 0;

      FIND_SOCKET
      
      /* Process Close */
      if (socket_connection != 0) {
	rval = socket_connection->Close(pid);
	if (rval == 1) {
	  iter.destroy();
	}
      }
      
      /* Return Result */
      ipc_connection->Reset();
      ipc_connection->S() << encode(rval);
      ipc_connection->Issue();
    } break;
  case CMD_BEGIN_MESSAGE:
    {
      FIND_SOCKET
      
      /* Process Begin Message */
      if (socket_connection != 0) {
	socket_connection->BeginMessage();
      }

      /* Return */
      ipc_connection->Reset();
      ipc_connection->S().put(CMD_MESSAGE_ACK);
      ipc_connection->Issue();
    } break;
  case CMD_END_MESSAGE:
    {
      FIND_SOCKET
      
      /* Process Begin Message */
      if (socket_connection != 0) {
	socket_connection->HandleData();
	socket_connection->EndMessage();
      }

      /* Return */
      ipc_connection->Reset();
      ipc_connection->S().put(CMD_MESSAGE_ACK);
      ipc_connection->Issue();
    } break;
  case CMD_GET_SOCK_NAME:
    {
      /* Return Value */
      int rval = 1;
      
      FIND_SOCKET
      
      /* Return Result */
      ipc_connection->Reset();
      /* Process GetSockName */
      if (socket_connection != 0) {
	socket_connection->GetSockName(ipc_connection->S());
      } else {
	ipc_connection->S() << encode(rval);
      }
      ipc_connection->Issue();
    } break;
  case CMD_SET_SOCK_OPT:
    {
      /* Return Value */
      int rval = -1;
      
      FIND_SOCKET

      /* Process SetSockOpt */
      if (socket_connection != 0) {
	rval = socket_connection->SetSockOpt(*msg);
      }

      /* Return Result */
      ipc_connection->Reset();
      ipc_connection->S() << encode(rval);
      ipc_connection->Issue();
    } break;
  case CMD_GET_SOCK_OPT:
    {
      /* Return Value */
      int rval = -1;
      
      FIND_SOCKET
      
      /* Return Result */
      ipc_connection->Reset();
      /* Process GetSockOpt */
      if (socket_connection != 0) {
	socket_connection->GetSockOpt(*msg,ipc_connection->S());
      } else {
	ipc_connection->S() << encode(rval);
      }
      ipc_connection->Issue();
    } break;
  case CMD_ACCEPT:
    {
      /* Return Value */
      int rval = -1;
      
      FIND_SOCKET
      
      /* Process Accept. */
      if (socket_connection != 0) {
	socket_connection->Accept(ipc_connection);
	break; /* Accept causes client to block on second call. */
      }
      
      /* Return Result */
      ipc_connection->Reset();
      ipc_connection->S() << encode(rval);
      ipc_connection->Issue();
    } break;
  case CMD_FORK:
    {
      /* Get Args */
      int pid;
      int ppid;
      (*msg) >> decode(pid);
      (*msg) >> decode(ppid);

      /* Find Sockets */
      MTCP_SOCKET* socket_connection;
      SLIST_ITERATOR iter(socket_connections);
      while ( socket_connection = (MTCP_SOCKET*)++iter ) {
	if (socket_connection->PIDEQ(ppid)) {
	  socket_connection->AppendPID(pid);
	}
      }
      ipc_connection->Reset();
      ipc_connection->S().put(CMD_FORK_ACK);
      ipc_connection->Issue();
    } break;
  }
  delete msg;
}

void
MTCP_SERVER::Handle_SocketData(CLIENTDATA _mtcp_socket) {
  MTCP_SOCKET* mtcp_socket = (MTCP_SOCKET*)_mtcp_socket;
  mtcp_socket->HandleData();
}

void
MTCP_SERVER::Handle_RMP_Event(CLIENTDATA _E) {
  mRMPEvent* E = (mRMPEvent*)_E;
  switch(E->returnRMPEventType()) {
  case MESSAGE:
    {
      int *fdList;
      if (E->num_fd() > 0) {
	fdList = E->client_fd();
	mMessage *mmsg = E->returnData();
	for (int i=0; i<mmsg->returnNumOfBuffers(); i++) {
	  char *msg_data = mmsg->returnNextMessage();
	  int msg_length = mmsg->returnNextMessageLength();
	  for (int j=0; j < E->num_fd(); j++) {
	    MTCP_SOCKET* socket_connection = (MTCP_SOCKET*)fdList[j];
	    socket_connection->HandleMessage(msg_data, msg_length);
	  }
	}
      }
    } break;
  case FORMED_OWN_GROUP:
    cerr << "Formed own group." << endl; cerr.flush();
    break;
  case JOINED_GROUP:
  case FAILURE_RECOVERY:
  case MEMBERSHIP_CHANGE:
    {
      MTCP_SOCKET* socket_connection;
      SLIST_ITERATOR iter(socket_connections);
      while( socket_connection = (MTCP_SOCKET*)++iter ) {
	socket_connection->Member(E->returnmRMPGroupForEvent());
      }
      cerr << "Group list updated." << endl; cerr.flush();
    } break;
  case LEFT_GROUP:
    cerr << "RMP: LEFT GROUP" << endl; cerr.flush();
    E->returnRMPGroupForEvent()->release();
    break;
  case FAILED_RECOVERY:
    cerr << "Failed Reformation." << endl;
    cerr.flush();
    break;
  case ATOMICITY_VIOLATED:
    cerr << "Atomicity Violated." << endl;
    cerr.flush();
    break;
#ifdef REFORM_DROPPED
  case REFORM_DROPPED:
    cerr << "Dropped from Reformation." << endl;
    cerr.flush();
    break;
#endif
  case LEFT_ALL_GROUPS:
    cerr << "RMP: LEFT ALL GROUPS" << endl; cerr.flush();
    if (terminate == 1) {
      exit(exit_status);
    }
    break;
  default:
    break;
  }
  E->release();
}

void
MTCP_SERVER::InitSignalHandlers() {
#ifdef SIGCHLD
  if (SIGNAL_INTERRUPT(SIGCHLD, MTCP_SERVER::SignalCHLD) == SIG_ERR) return;
#ifdef SIGABRT
  SIGNAL_INTERRUPT(SIGABRT, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGBUS
  SIGNAL_INTERRUPT(SIGBUS, MTCP_SERVER::ForwardSignal);
#endif
  signal_CONT = SIGNAL_INTERRUPT(SIGCONT, MTCP_SERVER::ForwardSignal);
#ifdef SIGEMT
  SIGNAL_INTERRUPT(SIGEMT, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGFPE
  SIGNAL_INTERRUPT(SIGFPE, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGHUP
  SIGNAL_INTERRUPT(SIGHUP, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGILL
  SIGNAL_INTERRUPT(SIGILL, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGINFO
  SIGNAL_INTERRUPT(SIGINFO, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGINT
  SIGNAL_INTERRUPT(SIGINT, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGIO
  SIGNAL_INTERRUPT(SIGIO, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGIOT
  SIGNAL_INTERRUPT(SIGIOT, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGPROF
  SIGNAL_INTERRUPT(SIGPROF, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGQUIT
  SIGNAL_INTERRUPT(SIGQUIT, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGSEGV
  SIGNAL_INTERRUPT(SIGSEGV, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGSYS
  SIGNAL_INTERRUPT(SIGSYS, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGTERM
  SIGNAL_INTERRUPT(SIGTERM, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGTRAP
  SIGNAL_INTERRUPT(SIGTRAP, MTCP_SERVER::ForwardSignal);
#endif
#ifdef ENABLE_FORWARD_SIGTSTP
  signal_TSTP = SIGNAL_INTERRUPT(SIGTSTP, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGTTIN
  SIGNAL_INTERRUPT(SIGTTIN, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGTTOU
  SIGNAL_INTERRUPT(SIGTTOU, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGUSR1
  SIGNAL_INTERRUPT(SIGUSR1, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGUSR2
  SIGNAL_INTERRUPT(SIGUSR2, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGVTALRM
  SIGNAL_INTERRUPT(SIGVTALRM, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGWINCH
  SIGNAL_INTERRUPT(SIGWINCH, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGXCPU
  SIGNAL_INTERRUPT(SIGXCPU, MTCP_SERVER::ForwardSignal);
#endif
#ifdef SIGXFSZ
  SIGNAL_INTERRUPT(SIGXFSZ, MTCP_SERVER::ForwardSignal);
#endif
#endif
}

void
MTCP_SERVER::ForwardSignal(int signo) {
  kill(IPC::PID(), signo);
  if (signo == SIGCONT) {
    signal_CONT(signo);
  }
}

void
MTCP_SERVER::SignalCHLD(int) {
  int stat;
  int options = 0;
  pid_t pid;
  
#ifdef WUNTRACED
  options |= WUNTRACED;
#endif
  pid = waitpid(-1, &stat, options);

  if (WIFEXITED(stat)) { /* Normal termination. */
    terminate = 1;
    exit_status = WEXITSTATUS(stat);
  } else if (WIFSIGNALED(stat)) { /* Abnormal termination. */
    terminate = 1;
    exit_status = 1;
  } else if (WIFSTOPPED(stat)) { /* Child stopped. */
#ifdef ENABLE_FORWARD_SIGTSTP
    signal_TSTP(WSTOPSIG(stat));
#endif
  }
}

int
MTCP_SERVER::TerminateQ() {
  static int terminated = 0;
  if (terminate == 1) {
    socket_connections.destroyAll();
    terminated = 1;
  }
  return terminate;
}
